/*
 * AbstractIntegerNumber.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;

/**
 * Abstract base class for <code>IntegerNumber</code> implementations.
 *
 * <p>Concrete subclasses must at least provide implementations for the
 * following methods:
 *
 * <ul>
 *    <li><code>getSign()</code> --
 *        Determines the sign of this number</li>
 *
 *    <li><code>negate()</code> --
 *        Computes -this</li>
 *
 *    <li><code>add(IntegerNumber)</code> --
 *        Computes this+<em>n</em>,
 *        where <em>n</em> is an integer number</li>
 *
 *    <li><code>multiply(IntegerNumber)</code> --
 *        Computes this*<em>n</em>,
 *        where <em>n</em> is an integer number</li>
 *
 *    <li><code>integerDivide(IntegerNumber)</code> --
 *        Computes the integer result of a division</li>
 *
 *    <li><code>longValue()</code> --
 *        Converts the value of this number to a <code>long</code></li>
 *
 *    <li><code>doubleValue()</code> --
 *        Converts the value of this number to a <code>double</code></li>
 *
 *    <li><code>toByteArray()</code> --
 *        Converts the value of this number to a set of bytes
 *        that represent the two's complement of this number</li>
 *
 * </ul>
 *
 * Subclasses are encouraged to override as many methods as they can in order
 * to optimize the implementation. Those methods marked as <code>final</code>
 * need no further optimization.
 *
 * @version 1.2 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */
public abstract class AbstractIntegerNumber
   extends AbstractRationalNumber
   implements IntegerNumber
{
   //=========================================================================
   // CONSTRUCTOR
   //=========================================================================

   /**
    * Constructs a new <code>AbstractIntegerNumber</code>.
    */
   protected AbstractIntegerNumber()
   {
      // empty
   }


   //=========================================================================
   // COMPARISON
   //=========================================================================

   public int compareTo(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      //-----------------------
      // PENDING: Optimize this
      //-----------------------

      byte[] a = toByteArray();
      byte[] b = n.toByteArray();

      return new BigInteger(a).compareTo(new BigInteger(b));
   }

   //=========================================================================
   // HASH CODE:
   //
   // - hashCode
   //
   //=========================================================================

   public int hashCode()
   {
      //-------------------------
      // NOTE: not very efficient
      //-------------------------

      return new Long(longValue()).hashCode();
   }


   //=========================================================================
   // PROPERTY: sign
   //
   // - getSign
   //
   //=========================================================================

   /**
    * Determines the sign of this number. This method returns -1, 0, or
    * 1 as this number is negative, zero, or positive.
    *
    * <p>The declaration is made abstract, because the implementation in
    * superclass <code>AbstractRationalNumber</code> does not make sense
    * for <code>IntegerNumber</code> implementations.
    *
    * @return
    *    the sign of this number.
    */
   public abstract int getSign();


   //=========================================================================
   // UNARY OPERATIONS:
   //
   // - negate
   // - invert
   //
   //=========================================================================

   /**
    * Computes -this.
    *
    * <p>The declaration is made abstract, because the implementation in
    * superclass <code>AbstractRationalNumber</code> does not make sense
    * for <code>IntegerNumber</code> implementations.
    *
    * @return
    *    the negative of this.
    */
   public abstract RealNumber negate();

   public RealNumber invert()
      throws ArithmeticException
   {
      //--------------------------------------------------------------------
      // NOTE: createFraction will check that the value of this is not zero.
      //--------------------------------------------------------------------

      return NumberCentral.createFraction(LongIntegerNumber.ONE, this);
   }


   //=========================================================================
   // BINARY OPERATIONS:
   //
   // - add
   // - subtract (2x)
   // - multiply
   // - divide
   // - pow
   //
   //=========================================================================

   public RationalNumber add(RationalNumber n)
      throws IllegalArgumentException
   {
      //-------------------------------------------------------
      // NOTE: null instanceof whatever   is never true, is it?
      //-------------------------------------------------------

      if (n instanceof IntegerNumber)
      {
         return ((IntegerNumber) this).add((IntegerNumber) n);
      }

      ExceptionSupport.checkNNotNull(n);

      IntegerNumber denominator = n.getDenominator();
      IntegerNumber numerator =
         ((IntegerNumber) this).multiply(denominator).add(n.getNumerator());

      return NumberCentral.createFraction(numerator, denominator);
   }

   public RationalNumber subtract(RationalNumber n)
      throws IllegalArgumentException
   {
      //-------------------------------------------------------
      // NOTE: null instanceof whatever   is never true, is it?
      //-------------------------------------------------------

      if (n instanceof IntegerNumber)
      {
         return subtract((IntegerNumber) n);
      }

      ExceptionSupport.checkNNotNull(n);

      IntegerNumber denominator = n.getDenominator();
      IntegerNumber numerator =
         ((IntegerNumber) this).multiply(denominator).subtract(
         n.getNumerator());

      return NumberCentral.createFraction(numerator, denominator);
   }

   public IntegerNumber subtract(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      IntegerNumber negated = (IntegerNumber) n.negate();
      return ((IntegerNumber) this).add(negated);
   }

   public RationalNumber multiply(RationalNumber n)
      throws IllegalArgumentException
   {
      if (n instanceof IntegerNumber)
      {
         return ((IntegerNumber) this).multiply((IntegerNumber) n);
      }

      ExceptionSupport.checkNNotNull(n);

      IntegerNumber numerator =
         ((IntegerNumber) this).multiply(n.getNumerator());
      IntegerNumber denominator = n.getDenominator();

      return NumberCentral.createFraction(numerator, denominator);
   }

   public RationalNumber divide(RationalNumber n)
      throws IllegalArgumentException, ArithmeticException
   {
      ExceptionSupport.checkNNotNull(n);

      IntegerNumber numerator =
         ((IntegerNumber) this).multiply(n.getDenominator());
      IntegerNumber denominator = n.getNumerator();

      return NumberCentral.createFraction(numerator, denominator);
   }

   public RationalNumber pow(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      if (n.getSign() == 1)
      {
         IntegerNumber result = this;
         for (n=n.dec(); n.getSign()==1; n=n.dec())
         {
            result = ((IntegerNumber) this).multiply(result);
         }
         return result;
      }

      // x**0 is allways 1
      if (n.equals(LongIntegerNumber.ZERO))
      {
         return LongIntegerNumber.ONE;
      }

      // n must be < 0

      // x ** n  (n<0) equals 1/(x ** -n)
      IntegerNumber absolute = (IntegerNumber) n.negate();
      RationalNumber invertedResult = pow(absolute);
      return (RationalNumber) (invertedResult.invert());
   }


   //=========================================================================
   // NUMERATOR AND DENOMINATOR:
   //
   // - getNumerator
   // - getDenominator
   //
   //=========================================================================

   /**
    * Returns the numerator of this rational number. For an integer number,
    * it is this, the number itself.
    *
    * @return
    *    the numerator, being this.
    */
   public final IntegerNumber getNumerator()
   {
      return this;
   }

   /**
    * Returns the denominator of this rational number. For an integer number,
    * this is allways 1.
    *
    * @return
    *    the denominator, being 1.
    */
   public final IntegerNumber getDenominator()
   {
      return LongIntegerNumber.ONE;
   }


   //=========================================================================
   // UNARY INTEGER OPERATIONS:
   //
   // - dec
   // - inc
   // - fac
   //
   //=========================================================================

   public IntegerNumber dec()
   {
      return subtract(LongIntegerNumber.ONE);
   }

   public IntegerNumber inc()
   {
      return ((IntegerNumber) this).add(LongIntegerNumber.ONE);
   }

   public IntegerNumber fac()
      throws IllegalArgumentException
   {
      if (getSign()==0)
      {
         return LongIntegerNumber.ZERO;
      }

      IntegerNumber one    = LongIntegerNumber.ONE;
      IntegerNumber result = this;
      IntegerNumber n = this.dec();
      for (; n.compareTo(one)==1; n=n.dec())
      {
         result = result.multiply(n);
      }
      return result;
   }


   //=========================================================================
   // BINARY INTEGER OPERATIONS:
   //
   // - remainder
   // - gcd
   // - isRelativePrime
   //
   //=========================================================================

   public IntegerNumber remainder(IntegerNumber n)
      throws ArithmeticException, IllegalArgumentException
   {
      //----------------------------------------
      // NOTE: subtract will check n is not null
      //----------------------------------------

      return (IntegerNumber) subtract(integerDivide(n).multiply(n)).abs();
   }

   public IntegerNumber gcd(IntegerNumber n)
      throws IllegalArgumentException
   {
      IntegerNumber u, temp;

      ExceptionSupport.checkNNotNull(n);

      n = (IntegerNumber) n.abs();
      u = (IntegerNumber) abs();

      // check argument is zero
      if (n.getSign() == 0)
      {
         return u;
      }

      // check this is zero
      if (u.getSign() == 0)
      {
         // THIS WAS: return (IntegerNumber) n.abs();
         return n;
      }

      while (u.getSign() == 1)
      {
         if (u.compareTo(n) == -1)
         {
            temp = u;
            u = n;
            n = temp;
         }
         u = u.remainder(n);
      }

      return n;
   }

   public boolean isRelativePrime(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      return (gcd(n).compareTo(LongIntegerNumber.ONE) == 0);
   }


   //=========================================================================
   // TEXTUAL CONVERSION:
   //
   // - toString
   //
   //=========================================================================

   /**
    * Returns a textual representation of this object.
    *
    * @return
    *    a <code>String</code> of this object.
    */
   public String toString()
   {
      //------------------------------
      // TODO: Have own implementation
      //------------------------------

      return new BigInteger(toByteArray()).toString();
   }

 
   //=========================================================================
   // NUMERIC CONVERSION:
   //
   // - toBigInteger
   // - toBigDecimal
   //
   //=========================================================================

   /**
    * Converts the value of this number to a <code>BigInteger</code>.
    *
    * @return
    *    a <code>BigInteger</code> with the value of this.
    */
   public BigInteger toBigInteger()
   {
      return new BigInteger(toByteArray());
   }

   /**
    * Converts the value of this number to a <code>BigDecimal</code>. A
    * precision nor a rounding mode need to be specified for integer numbers.
    *
    * @return
    *    a <code>BigDecimal</code> with the exact value of this.
    */
   public BigDecimal toBigDecimal()
   {
      return new BigDecimal(toBigInteger());
   }

   /**
    * Converts the value of this number to a <code>BigDecimal</code> with the
    * specified precision. The <em>precision</em> argument is ignored,
    * because no rounding is needed.
    *
    * @param precision
    *    the number of digits behind the decimal point.
    *
    * @return
    *    a <code>BigDecimal</code> with the exact value of this.
    *
    * @exception IllegalArgumentException
    *    if <em>precision</em> &lt; 0
    */
   public BigDecimal toBigDecimal(int precision)
      throws IllegalArgumentException
   {
      //------------------------
      // PENDING: Look into this
      //------------------------

      return toBigDecimal();
   }

   /**
    * Converts the value of this number to a <code>BigDecimal</code> with the
    * specified precision and rounding mode. The <em>precision</em> and
    * <em>roundingMode</em> arguments are ignored, because no rounding is
    * needed.
    *
    * @param precision
    *    the number of digits behind the decimal point.
    *
    * @param roundingMode
    *    the rounding mode to use, one of the modes defined
    *    in class <code>BigDecimal</code>.
    *
    * @return
    *    a <code>BigDecimal</code> with the rounded value of this.
    *
    * @exception IllegalArgumentException
    *    if one of the following applies:
    *    <ol>
    *       <li><em>precision</em> &lt; 0
    *       <li><em>roundingMode</em> does not have a valid value
    *    </ol>
    */
   public BigDecimal toBigDecimal(int precision, int roundingMode)
   {
      //------------------------
      // PENDING: Look into this
      //------------------------

      return toBigDecimal();
   }


   //=========================================================================
   // ROUNDING:
   //
   // - trunc
   //
   //=========================================================================

   /**
    * Rounds to an integer number towards 0.
    *
    * <p>The implementation of this method in
    * <code>AbstractIntegerNumber</code> returns this.
    *
    * @return
    *    this number truncated to an integer.
    */
   public final IntegerNumber trunc()
   {
      return this;
   }

}
