/*
 * AbstractRationalNumber.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;

/**
 * Abstract base class for <code>RationalNumber</code> implementations.
 *
 * <p>Concrete subclasses should at least provide an implementations for the
 * following methods:
 *
 * <ul>
 *    <li><code>getNumerator()</code> --
 *        Returns the numerator of this fraction</li>
 *
 *    <li><code>getDenominator()</code> --
 *        Returns the denominator of this fraction</li>
 * </ul>
 *
 * <code>IntegerNumber</code> implementations based on this class must
 * override the following methods:
 *
 * <ul>
 *    <li><code>getSign()</code> --
 *        The implementation in this class returns
 *        the sign of the numerator, which is <code>this</code> in an
 *        <code>IntegerNumber</code> implementation</li>
 *
 *    <li><code>negate()</code> --
 *        The implementation in this class returns a
 *        fraction with a negated numerator, which is <code>this</code> in an
 *        <code>IntegerNumber</code> implementation</li>
 *
 *    <li><code>toString()</code> --
 *        The implementation in this class
 *        returns a string representation of the numerator and the string
 *        representation of the denominator; in an <code>IntegerNumber</code>
 *        implementation the numerator is this</li>
 *
 * </ul>
 *
 * @version 1.2 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */
public abstract class AbstractRationalNumber
   extends AbstractRealNumber
   implements RationalNumber
{
   //=========================================================================
   // CONSTRUCTOR
   //=========================================================================

   /**
    * Constructs a new <code>AbstractRationalNumber</code>.
    */
   protected AbstractRationalNumber()
   {
      // empty
   }


   //=========================================================================
   // COMPARISON
   //=========================================================================

   public int compareTo(RationalNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      int thisSign = getSign();
      int thatSign = n.getSign();

      if (thisSign != thatSign)
      {
         if (thisSign > thatSign) return  1;
         if (thisSign < thatSign) return -1;
         return 0;
      }

      return subtract(n).getSign();
   }


   //=========================================================================
   // PROPERTY: sign
   //
   // - getSign
   //
   //=========================================================================

   public int getSign()
   {
      return getNumerator().getSign();
   }


   //=========================================================================
   // COMPARISON:
   //
   // - retryCompare
   //
   //=========================================================================

   public int retryCompare(RealNumber n)
      throws IllegalArgumentException,
             CanNotCompareException
   {
      //-------------------------------------------------------
      // NOTE: null instanceof whatever   is never true, is it?
      //-------------------------------------------------------

      //-------------------------------------------------------
      // NOTE: the other retryCompare() methods will check n!=null
      //-------------------------------------------------------

      if (n instanceof RationalNumber)
      {
         return retryCompare((RationalNumber) n);
      }
      return super.retryCompare(n);
   }

 
   public int retryCompare(RationalNumber n)
      throws IllegalArgumentException
   {
      //----------------------------------------
      // NOTE: subtract() will check (n != null)
      //----------------------------------------

      return (subtract(n).getSign());
   }

 
   //=========================================================================
   // UNARY OPERATIONS:
   //
   // - negate
   // - invert
   //
   //=========================================================================

   public RealNumber negate()
   {
      return NumberCentral.createFraction(
         (IntegerNumber) getNumerator().negate(), getDenominator());
   }

   public RealNumber invert()
   {
      return NumberCentral.createFraction(getDenominator(), getNumerator());
   }


   //=========================================================================
   // BINARY OPERATIONS:
   //
   // - add (2x)
   // - subtract (2x)
   // - multiply (2x)
   // - divide (2x)
   // - pow (2x)
   //
   //=========================================================================

   public RealNumber add(RealNumber n)
   {
      //-------------------------------------------------------
      // NOTE: null instanceof whatever   is never true, is it?
      //-------------------------------------------------------

      //-------------------------------------------------
      // NOTE: the other add() methods will check n!=null
      //-------------------------------------------------

      if (n instanceof RationalNumber)
      {
         return add((RationalNumber) n);
      }
      return super.add(n);
   }

   public RationalNumber add(RationalNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      IntegerNumber a = getNumerator().multiply(n.getDenominator());
      IntegerNumber b = getDenominator().multiply(n.getNumerator());

      IntegerNumber numerator = a.add(b);
      IntegerNumber denominator =
         getDenominator().multiply(n.getDenominator());

      return NumberCentral.createFraction(numerator, denominator);
   }

   public RealNumber subtract(RealNumber n)
      throws IllegalArgumentException
   {
      //-------------------------------------------------------
      // NOTE: null instanceof whatever   is never true, is it?
      //-------------------------------------------------------

      //------------------------------------------------------
      // NOTE: the other subtract() methods will check n!=null
      //------------------------------------------------------

      if (n instanceof RationalNumber)
      {
         return subtract((RationalNumber) n);
      }
      return super.subtract(n);
   }

   public RationalNumber subtract(RationalNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      IntegerNumber a = getNumerator().multiply(n.getDenominator());
      IntegerNumber b = getDenominator().multiply(n.getNumerator());

      IntegerNumber numerator = a.subtract(b);
      IntegerNumber denominator =
         getDenominator().multiply(n.getDenominator());

      return NumberCentral.createFraction(numerator, denominator);
   }

   public RealNumber multiply(RealNumber n)
   {
      //-------------------------------------------------------
      // NOTE: null instanceof whatever   is never true, is it?
      //-------------------------------------------------------

      //------------------------------------------------------
      // NOTE: the other multiply() methods will check n!=null
      //------------------------------------------------------

      if (n instanceof RationalNumber)
      {
         return multiply((RationalNumber) n);
      }
      return super.multiply(n);
   }

   public RationalNumber multiply(RationalNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      IntegerNumber numerator = getNumerator().multiply(n.getNumerator());
      IntegerNumber denominator =
         getDenominator().multiply(n.getDenominator());

      return NumberCentral.createFraction(numerator, denominator);
   }

   public RealNumber divide(RealNumber n)
      throws IllegalArgumentException, ArithmeticException
   {
      //-------------------------------------------------------
      // NOTE: null instanceof whatever   is never true, is it?
      //-------------------------------------------------------

      //----------------------------------------------------
      // NOTE: the other divide() methods will check n!=null
      //----------------------------------------------------

      if (n instanceof RationalNumber)
      {
         return divide((RationalNumber) n);
      }
      return super.divide(n);
   }

   public RationalNumber divide(RationalNumber n)
      throws ArithmeticException,
             IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      IntegerNumber numerator = getNumerator().multiply(n.getDenominator());
      IntegerNumber denominator = getDenominator().multiply(n.getNumerator());

      return NumberCentral.createFraction(numerator, denominator);
   }

   public RealNumber pow(RealNumber n)
      throws IllegalArgumentException
   {
      //-------------------------------------------------------
      // NOTE: null instanceof whatever   is never true, is it?
      //-------------------------------------------------------

      //----------------------------------------------------
      // NOTE: the other pow() methods will check n!=null
      //----------------------------------------------------

      if (n instanceof IntegerNumber)
      {
         return pow((IntegerNumber) n);
      }
      return super.pow(n);
   }

   public RationalNumber pow(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      if (n.getSign() == 1)
      {
         RationalNumber result = this;
         for (n=n.dec(); n.getSign()==1; n=n.dec())
         {
            result = multiply(result);
         }
         return result;
      }

      // x**0 is allways 1
      if (n.equals(LongIntegerNumber.ZERO))
      {
         return LongIntegerNumber.ONE;
      }

      // n must be < 0

      // x ** n  (n<0) equals 1/(x ** -n)
      IntegerNumber absolute = (IntegerNumber) n.negate();
      RationalNumber invertedResult = pow(absolute);
      return (RationalNumber) (invertedResult.invert());
   }


   //=========================================================================
   // TEXTUAL CONVERSION:
   //
   // - toString()
   //
   //=========================================================================

   public String toString()
   {
      //------------------------------------------------------
      // TO DO: Look into performance. Maybe use StringBuffer?
      //------------------------------------------------------

      return ""+getNumerator()+"/"+getDenominator();
   }


   //=========================================================================
   // NUMERIC CONVERSION:
   //
   // - doubleValue
   // - toBigDecimal (2x)
   //
   //=========================================================================

   public double doubleValue()
   {
      return getNumerator().doubleValue() / getDenominator().doubleValue();
   }

   public BigDecimal toBigDecimal(int precision)
      throws IllegalArgumentException
   {
      //---------------------------------------------------------
      // TO DO: Check that the IllegalArgumentException is thrown
      //        as specified in the interface contract
      //---------------------------------------------------------

      return toBigDecimal(precision, BigDecimal.ROUND_HALF_UP);
   }

   public BigDecimal toBigDecimal(int precision, int roundingMode)
      throws IllegalArgumentException
   {
      //---------------------------------------------------------
      // TO DO: Check that the IllegalArgumentException is thrown
      //        as specified in the interface contract
      //---------------------------------------------------------

      BigDecimal numerator = getNumerator().toBigDecimal();
      BigDecimal denominator = getDenominator().toBigDecimal();

      return
         numerator.divide(denominator, precision, roundingMode);
   }


   //=========================================================================
   // ROUNDING:
   //
   // - trunc
   //
   //=========================================================================

   public IntegerNumber trunc()
   {
      return getNumerator().integerDivide(getDenominator());
   }

}
