/*
 * AbstractRealNumber.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;

/**
 * Abstract base class for <code>RealNumber</code> implementations.
 *
 * <p>Concrete subclasses should provide implementations for the following
 * methods:
 *
 * <ul>
 *    <li><code>getSign()</code> --
 *        Determines the sign of this number</li>
 *
 *    <li><code>toBigDecimal(int,int)</code> --
 *        Converts the value of this
 *        number to a <code>BigDecimal</code> with the specified precision,
 *        using the specified rounding mode</li>
 *
 *    <li><code>trunc()</code> --
 *        Truncates the value of this number to
 *        an <code>IntegerNumber</code></li>
 *
 * </ul>
 *
 * @version 1.2 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */
public abstract class AbstractRealNumber
   extends    Number
   implements RealNumber
{
   //=========================================================================
   // CONSTRUCTOR
   //=========================================================================

   /**
    * Constructs a new <code>AbstractRealNumber</code> object.
    */
   protected AbstractRealNumber()
   {
      // empty
   }


   //=========================================================================
   // HASH CODE
   //=========================================================================

   public int hashCode()
   {
      return new Double(doubleValue()).hashCode();
   }

   public final boolean equals(Object o)
   {
      if (o == null)
      {
         return false;
      }

      if (o instanceof RealNumber)
      {
         RealNumber n = (RealNumber) o;

         try
         {
            return retryCompare(n) == 0;
         }
         catch (CanNotCompareException e)
         {
            //----------------------------------------------------------
            // PENDING: modify when retryCompare(RealNumber, int digits)
            //          (or something alike) is implemented
            //----------------------------------------------------------

            return false;
         }
      }

      return false;
   }


   //=========================================================================
   // COMPARISON:
   //
   // - compareTo
   // - retryCompare
   //
   //=========================================================================

   public final int compareTo(RealNumber n)
      throws IllegalArgumentException,
             CanNotCompareException
   {
      try
      {
         return retryCompare(n);
      }
      catch(CanNotCompareException e)
      {
         return -n.retryCompare(this);
      }
   }

   public int retryCompare(RealNumber n)
      throws IllegalArgumentException,
             CanNotCompareException
   {
      //----------------------------------------
      // NOTE: subtract() will check (n != null)
      //----------------------------------------

      return (subtract(n).getSign());
   }


   //=========================================================================
   // UNARY OPERATIONS:
   //
   // - abs
   // - negate
   // - invert
   //
   //=========================================================================

   public RealNumber abs()
   {
      if (getSign() == -1)
      {
         return negate();
      }
      return this;
   }

   public RealNumber negate()
   {
      return multiply(NumberCentral.valueOf(-1));
   }

   public RealNumber invert()
   {
      return pow(NumberCentral.valueOf(-1));
   }


   //=========================================================================
   // BINARY OPERATIONS:
   //
   // - add
   // - subtract
   // - multiply
   // - divide
   // - pow
   //
   //=========================================================================

   public RealNumber add(RealNumber n)
      throws IllegalArgumentException
   {
      //-----------------------------------------------
      // NOTE: NumberCentral.add will check (n != null)
      //-----------------------------------------------

      return NumberCentral.add(this, n);
   }

   public RealNumber subtract(RealNumber n)
      throws IllegalArgumentException
   {
      //--------------------------------------------
      // NOTE: the add method will check (n != null)
      //--------------------------------------------

      return add(n.negate());
   }

   public RealNumber multiply(RealNumber n)
   {
      //----------------------------------------------------
      // NOTE: NumberCentral.multiply will check (n != null)
      //----------------------------------------------------

      return NumberCentral.multiply(this, n);
   }

   public RealNumber divide(RealNumber n)
      throws IllegalArgumentException, ArithmeticException
   {
      //-------------------------------------------------
      // NOTE: the multiply method will check (n != null)
      //-------------------------------------------------

      //--------------------------------------------------------
      // NOTE: the invert method will check (value of n != zero)
      //--------------------------------------------------------

      return multiply(n.invert());
   }

   public RealNumber pow(RealNumber n)
      throws IllegalArgumentException
   {
      //-----------------------------------------------
      // NOTE: NumberCentral.pow will check (n != null)
      //-----------------------------------------------

      return NumberCentral.pow(this, n);
   }


   //=========================================================================
   // NUMERIC CONVERSION FIT CHECKING:
   //
   // - fitsLong
   // - fitsInt
   // - fitsShort
   // - fitsByte
   // - fitsDouble
   // - fitsFloat
   //
   //=========================================================================

   public boolean fitsLong()
   {
      byte[] byteArray = trunc().toByteArray();
      return (byteArray.length <= 8);
   }

   public boolean fitsInt()
   {
      byte[] byteArray = trunc().toByteArray();
      return (byteArray.length <= 4);
   }

   public boolean fitsShort()
   {
      byte[] byteArray = trunc().toByteArray();
      return (byteArray.length <= 2);
   }

   public boolean fitsByte()
   {
      byte[] byteArray = trunc().toByteArray();
      return (byteArray.length <= 1);
   }

   public boolean fitsDouble()
   {
      //-----------------
      // PENDING: Do this
      //-----------------

      return true;
   }

   public boolean fitsFloat()
   {
      //-----------------
      // PENDING: Do this
      //-----------------

      return true;
   }


   //=========================================================================
   // NUMERIC CONVERSION:
   //
   // - longValue
   // - intValue
   // - shortValue
   // - byteValue
   // - doubleValue
   // - floatValue
   // - toBigDecimal
   // - toBigInteger
   //
   //=========================================================================

   public long longValue()
   {
      return trunc().longValue();
   }

   public int intValue()
   {
      return (int) longValue();
   }

   public short shortValue()
   {
      return (short) longValue();
   }

   public byte byteValue()
   {
      return (byte) longValue();
   }

   public double doubleValue()
   {
      //----------------------------------------------------
      // PENDING: what precision should be used for doubles?
      //          30 is pretty safe for now
      //----------------------------------------------------

      return toBigDecimal(30).doubleValue();
   }

   public float floatValue()
   {
      return (float) doubleValue();
   }

   public BigDecimal toBigDecimal(int precision)
      throws IllegalArgumentException
   {
      return toBigDecimal(precision, BigDecimal.ROUND_HALF_UP);
   }

   public BigInteger toBigInteger()
   {
      //--------------------------
      // PENDING: Is this correct?
      //--------------------------

      return toBigDecimal(0).toBigInteger();
   }

}
