/*
 * BasicIntegerNumber.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;

/**
 * Basic implementation of an integer number. This implementation is based on
 * a <code>java.math.BigInteger</code>. It can hold all values from 
 * -2**32 to (2**32 - 1).
 *
 * <p>Instances must be obtained by using the
 * <code>createInstance()</code> factory method.
 *
 * @version 1.1 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 *
 * @see java.math.BigInteger
 */
public class BasicIntegerNumber
   extends AbstractIntegerNumber
{
   //=========================================================================
   // CONSTRUCTORS
   //=========================================================================

   /**
    * Constructs a <code>BasicIntegerNumber</code> from a
    * <code>BigInteger</code>.
    *
    * @param n the value for the new number.
    */
   protected BasicIntegerNumber(BigInteger n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      bigInteger = n;
   }


   //=========================================================================
   // FACTORY METHOD:
   //
   // - createInstance
   //
   //=========================================================================

   /**
    * Returns an instance of a <code>BasicIntegerNumber</code> based on a
    * <code>java.math.BigInteger</code>.
    *
    * @param n the long number to construct a
    *    <code>BasicIntegerNumber</code> from.
    * @return the (possibly newly constructed)
    *    <code>BasicIntegerNumber</code>.
    * @exception IllegalArgumentException if the argument is
    *    <code>null</code>.
    */
   public static BasicIntegerNumber createInstance(BigInteger n)
      throws IllegalArgumentException
   {
      //-----------------------------------------------
      // NOTE: constructor will make sure n is not null
      //-----------------------------------------------

      return new BasicIntegerNumber(n);
   }


   //=========================================================================
   // HASH CODE:
   //
   // - hashCode
   //
   //=========================================================================

   public int hashCode()
   {
      return bigInteger.hashCode();
   }


   //=========================================================================
   // PROPERTY: sign
   //=========================================================================

   public int getSign()
   {
      return bigInteger.signum();
   }


   //=========================================================================
   // UNARY OPERATIONS:
   //
   // - negate
   //
   //=========================================================================

   //------------------------------------------
   // METHOD OVERRIDDEN TO OPTIMIZE PERFORMANCE
   //------------------------------------------
   public RealNumber negate()
   {
      return createInstance(bigInteger.negate());
   }


   //=========================================================================
   // BINARY OPERATIONS:
   //
   // - add
   // - multiply
   //
   //=========================================================================

   public IntegerNumber add(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      BigInteger bigInt;
      if (n instanceof BasicIntegerNumber)
      {
         bigInt = ((BasicIntegerNumber) n).bigInteger;
      }
      else
      {
         bigInt = new BigInteger(n.toByteArray());
      }

      return createInstance(bigInteger.add(bigInt));
   }

   public IntegerNumber multiply(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      BigInteger bigInt;
      if (n instanceof BasicIntegerNumber)
      {
         bigInt = ((BasicIntegerNumber) n).bigInteger;
      }
      else
      {
         bigInt = new BigInteger(n.toByteArray());
      }

      return createInstance(bigInteger.multiply(bigInt));
   }


   //=========================================================================
   // SPECIFIC INTEGER OPERATIONS:
   //
   // - integerDivide
   //
   //=========================================================================

   public IntegerNumber integerDivide(IntegerNumber n)
      throws ArithmeticException, IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);
      ExceptionSupport.checkDivideByZero(n);

      BigInteger bigInt;
      if (n instanceof BasicIntegerNumber)
      {
         bigInt = ((BasicIntegerNumber) n).bigInteger;
      }
      else
      {
         bigInt = new BigInteger(n.toByteArray());
      }

      return createInstance(bigInteger.divide(bigInt));
   }

 
   //=========================================================================
   // NUMERIC CONVERSION:
   //
   // - longValue
   // - doubleValue
   // - toByteArray
   // - toBigInteger
   //
   //=========================================================================

   public long longValue()
   {
      return bigInteger.longValue();
   }

   public double doubleValue()
   {
      return bigInteger.doubleValue();
   }

   public byte[] toByteArray()
   {
      return bigInteger.toByteArray();
   }

   public BigInteger toBigInteger()
   {
      return bigInteger;
   }


   //=========================================================================
   // PROTECTED PROPERTY: bigInteger
   //=========================================================================

   /**
    * The <code>BigInteger</code> this
    * <code>IntegerNumber</code> implementation is based on.
    */
   protected BigInteger bigInteger;


}
