/*
 * DefaultNumberFactory.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;
import java.util.*;

/**
 * Default implementation of a <code>NumberFactory</code>.
 *
 * @version 1.2 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */
public class DefaultNumberFactory
   extends Object
   implements NumberFactory
{
   //=========================================================================
   // CONSTRUCTOR
   //=========================================================================

   /**
    * Constructs a new <code>DefaultNumberFactory</code>.
    */
   protected DefaultNumberFactory()
   {
      // empty
   }


   //=========================================================================
   // SHARED_INSTANCE
   //=========================================================================

   /**
    * The shared instance of this class. The instance can be shared
    * because it has no state.
    */
   public static DefaultNumberFactory SHARED_INSTANCE =
      new DefaultNumberFactory();


   //=========================================================================
   // FACTORY METHODS FOR COMPOSITES:
   //
   // - add
   // - multiply
   // - pow
   //
   //=========================================================================

   public RealNumber add(RealNumber a, RealNumber b)
      throws IllegalArgumentException
   {
      if (a instanceof RationalNumber && b instanceof RationalNumber)
      {
         return ((RationalNumber) a).add((RationalNumber) b);
      }

      // implicit else
      return Sum.createInstance(a, b);
   }

   public RealNumber multiply(RealNumber a, RealNumber b)
      throws IllegalArgumentException
   {
      if (a instanceof RationalNumber && b instanceof RationalNumber)
      {
         return ((RationalNumber) a).multiply((RationalNumber) b);
      }

      // implicit else
      return Product.createInstance(a, b);
   }

   public RealNumber pow(RealNumber base, RealNumber exponent)
      throws IllegalArgumentException, ArithmeticException
   {
      if (base     instanceof RationalNumber &&
          exponent instanceof IntegerNumber)
      {
         return ((RationalNumber) base).pow((IntegerNumber) exponent);
      }

      // implicit else
      return Power.createInstance(base, exponent);
   }


   //=========================================================================
   // FACTORY METHODS:
   //
   // - createFraction (2x)
   // - createInteger
   // - valueOf (4x)
   //
   //=========================================================================

   public RationalNumber createFraction(long numerator, long denominator)
      throws ArithmeticException
   {
      return createFraction(valueOf(numerator), valueOf(denominator));
   }

   public RationalNumber createFraction(IntegerNumber numerator,
                                        IntegerNumber denominator)

      throws IllegalArgumentException,
             ArithmeticException
   {
      ExceptionSupport.checkNumeratorNotNull(numerator);
      ExceptionSupport.checkDenominatorNotNull(denominator);
      ExceptionSupport.checkDivideByZero(denominator);

      // result of division may be integer
      IntegerNumber remainder = numerator.remainder(denominator);
      if (remainder.getSign() == 0)
      {
         return numerator.integerDivide(denominator);
      }

      //return BasicRationalNumber.createInstance(numerator, denominator);
      RationalNumber n =
         BasicRationalNumber.createInstance(numerator, denominator);


      return n;
   }

   public IntegerNumber createInteger(byte[] bytes)
   {
      //--------------------------------------
      // PENDING: Returns LongIntegerNumber if
      //          bytes.length <= 8
      //--------------------------------------

      return BasicIntegerNumber.createInstance(new BigInteger(bytes));
   }

   public RationalNumber valueOf(double n)
      throws IllegalArgumentException
   {
      // check argument is legal
      if (n==Double.POSITIVE_INFINITY ||
          n==Double.NEGATIVE_INFINITY ||
          n==Double.NaN)
      {
         throw new IllegalArgumentException(
            "argument cannot be Infinity, -Infinity or NaN");
      }

      long bits = Double.doubleToLongBits(n);

      if(bits == 0x0000000000000000L || bits == 0x8000000000000000L)
      {
         return LongIntegerNumber.ZERO;
      }

      long sign = bits >>> 63;
      long exp = ((bits >>> 52) & 0x7ff) - 1075;
      long mantissa = (bits & 0x000fffffffffffffL) + 0x0010000000000000L;

      // compute unsigned result
      IntegerNumber base = valueOf(mantissa);
      RationalNumber factor = valueOf(2).pow(valueOf(exp));

      RationalNumber result = base.multiply(factor);

      if(sign == 0)
      {
         return result;
      }
      else
      {
         return (RationalNumber) (result.negate());
      }
   }

   public IntegerNumber valueOf(long n)
   {
      return LongIntegerNumber.createInstance(n);
   }

   public IntegerNumber valueOf(BigInteger n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      return BasicIntegerNumber.createInstance(n);
   }

   public RationalNumber valueOf(BigDecimal n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      //----------------------------------------------------------------
      // NOTE: JavaSoft should rename scale() to getScale()
      //----------------------------------------------------------------

      int intScale = n.scale();

      //----------------------------------------------------------------
      // NOTE: JavaSoft should provide read access to the private intVal
      //       field!
      //----------------------------------------------------------------

      IntegerNumber numerator =
         valueOf(n.movePointRight(n.scale()).toBigInteger());


      IntegerNumber value10 = valueOf(10);
      IntegerNumber scale = valueOf(intScale);

      IntegerNumber denominator = (IntegerNumber) value10.pow(scale);

      return createFraction(numerator, denominator);
   }

   public RealNumber valueOf(String s)
      throws IllegalArgumentException,
             NumberFormatException
   {
      ExceptionSupport.checkSNotNull(s);

      //-----------------------------
      // TO DO: Implement this method
      //-----------------------------

      return null;
   }


   //=========================================================================
   // RANDOM NUMBERS:
   //
   // - randomGenerator
   // - createRandomInteger
   //
   //=========================================================================

   /**
    * The <code>Random</code> object used to obtain pseudo-random numbers.
    *
    * <p>This field is initialized upon construction.
    */
   protected Random randomGenerator = new Random();

   public IntegerNumber createRandomInteger(int maxBytes)
      throws IllegalArgumentException
   {
      if (maxBytes < 1)
      {
         throw new IllegalArgumentException("maxBytes < 1");
      }

      int n = (randomGenerator.nextInt() % maxBytes) + 1;
      n = n<0 ? -n : n;
      n %=  maxBytes;
      n++;
      byte[] array = new byte[n];
      randomGenerator.nextBytes(array);

      BigInteger bigInteger = new BigInteger(array);
      return BasicIntegerNumber.createInstance(bigInteger);
   }

}
