/*
 * ExceptionSupport.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

/**
 * This class provides exception throwing support for classes in this package.
 * Provided are:
 * <ul>
 *    <li><em>check methods</em> --
 *        check methods that throw an exception if some illegal condition is
 *        true
 *
 *    <li><em>shared message constants</em> --
 *        shared message constants to be used when throwing exceptions.
 *
 * </ul>
 *
 * @version 1.0 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */
public class ExceptionSupport
   extends Object
{

   //=========================================================================
   // CHECK METHODS:
   //
   // - checkANotNull
   // - checkBNotNull
   // - checkNNotNull (3x)
   // - checkSNotNull
   // - checkNumeratorNotNull
   // - checkDenominatorNotNull
   //
   // - checkDenominatorNotZero
   //
   //=========================================================================

   public static void checkANotNull(Object a)
      throws IllegalArgumentException
   {
      if (a == null)
      {
         throw new IllegalArgumentException(A_IS_NULL_MESSAGE);
      }
   }

   public static void checkBNotNull(Object b)
      throws IllegalArgumentException
   {
      if (b == null)
      {
         throw new IllegalArgumentException(B_IS_NULL_MESSAGE);
      }
   }

   public static void checkNNotNull(Object n)
      throws IllegalArgumentException
   {
      if (n == null)
      {
         throw new IllegalArgumentException(N_IS_NULL_MESSAGE);
      }
   }

   public static void checkSNotNull(Object s)
      throws IllegalArgumentException
   {
      if (s == null)
      {
         throw new IllegalArgumentException(S_IS_NULL_MESSAGE);
      }
   }

   public static void checkNumeratorNotNull(IntegerNumber numerator)
      throws IllegalArgumentException
   {
      if (numerator == null)
      {
         throw new IllegalArgumentException(NUMERATOR_IS_NULL_MESSAGE);
      }
   }

   public static void checkDenominatorNotNull(IntegerNumber demominator)
      throws IllegalArgumentException
   {
      if (demominator == null)
      {
         throw new IllegalArgumentException(DENOMINATOR_IS_NULL_MESSAGE);
      }
   }

   public static void checkOperandNotNull(RealNumber operand)
      throws IllegalArgumentException
   {
      if (operand == null)
      {
         throw new IllegalArgumentException(OPERAND_IS_NULL_MESSAGE);
      }
   }

   public static void checkDivideByZero(RealNumber n)
      throws ArithmeticException
   {
      if (n.getSign() == 0)
      {
         throw new ArithmeticException(DIVIDE_BY_ZERO_MESSAGE);
      }
   }


   //=========================================================================
   // MESSAGE STRING CONSTANTS
   //=========================================================================

   /**
    * The message for an exception when some parameter <em>a</em> is
    * <code>null</code>.
    */
   public final static String A_IS_NULL_MESSAGE = "a == null";

   /**
    * The message for an exception when some parameter <em>b</em> is
    * <code>null</code>.
    */
   public final static String B_IS_NULL_MESSAGE = "b == null";

   /**
    * The message for an exception when some parameter <em>n</em> is
    * <code>null</code>.
    */
   public final static String N_IS_NULL_MESSAGE = "n == null";

   /**
    * The message for an exception when some parameter <em>s</em> is
    * <code>null</code>.
    */
   public final static String S_IS_NULL_MESSAGE = "s == null";

   /**
    * The message for an exception when some parameter <em>operand</em> is
    * <code>null</code>.
    */
   public final static String OPERAND_IS_NULL_MESSAGE = "operand == null";

   /**
    * The message for an exception when some parameter <em>operands</em> is
    * <code>null</code>.
    */
   public final static String OPERANDS_IS_NULL_MESSAGE = "operands == null";

   /**
    * The message for an exception when some parameter <em>numerator</em> is
    * <code>null</code>.
    */
   public final static String NUMERATOR_IS_NULL_MESSAGE =
      "numerator == null";

   /**
    * The message for an exception when some parameter <em>denominator</em> is
    * <code>null</code>.
    */
   public final static String DENOMINATOR_IS_NULL_MESSAGE =
      "denominator == null";

   /**
    * The message for an exception thrown in case of a divide by zero.
    */
   public final static String DIVIDE_BY_ZERO_MESSAGE = "divide by zero";

}
