/*
 * LongIntegerNumber.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;

/**
 * Implementation of an <code>IntegerNumber</code> based on a
 * <code>long</code> value.
 *
 * <p>Instances must be obtained by using one of the
 * <code>createInstance()</code> factory methods. Using this approach,
 * instances can be transparently cached by this class.
 *
 * @version 1.1 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */

//---------------------------------------------------------------------------
// CHANGES SINCE 1.0a3:
//
//   And changed createInstance() to return one of these if appropriate
//
//---------------------------------------------------------------------------

public class LongIntegerNumber
   extends AbstractIntegerNumber
{
   //=========================================================================
   // CONSTRUCTORS
   //=========================================================================

   /**
    * Constructs a <code>LongIntegerNumber</code> from a
    * <code>long</code>.
    *
    * @param n
    *    the value for the new number.
    */
   protected LongIntegerNumber(long n)
   {
      value = n;
   }


   //=========================================================================
   // FACTORY METHODS:
   //
   // - createInstance(long)
   //
   //=========================================================================

   /**
    * Returns an instance of a <code>LongIntegerNumber</code> based on a
    * <code>long</code> value.
    *
    * @param n the long number to construct a
    *    <code>LongIntegerNumber</code> from.
    *
    * @return the (possibly newly constructed)
    *    <code>LongIntegerNumber</code>.
    */
   public static LongIntegerNumber createInstance(long n)
   {
      if (n>=-2 && n<=10)
      {
         switch ((int) n)
         {
            case  0: return ZERO;
            case  1: return ONE;
            case  2: return TWO;
            case -1: return MINUS_ONE;
            case -2: return MINUS_TWO;
            case  3: return THREE;
            case  4: return FOUR;
            case  5: return FIVE;
            case  6: return SIX;
            case  7: return SEVEN;
            case  8: return EIGHT;
            case  9: return NINE;
            case 10: return TEN;
         }
      }
      return new LongIntegerNumber(n);
   }


   //=========================================================================
   // CACHED INSTANCES:
   //
   // - MINIMUM_VALUE
   // - MAXIMUM_VALUE
   //
   // - MINUS_TWO
   // - MINUS_ONE
   // - ZERO
   // - ONE
   // - TWO
   // - THREE
   // - FOUR
   // - FIVE
   // - SIX
   // - SEVEN
   // - EIGHT
   // - NINE
   // - TEN
   //
   // These fields have been made public, because this will enable classes
   // outside this package instant access to these often-used instances.
   //=========================================================================

   /**
    * The minimum <code>IntegerNumber</code> that can be represented by this
    * class.
    */
   public final static LongIntegerNumber MINIMUM_VALUE =
      new LongIntegerNumber(Long.MIN_VALUE);

   /**
    * The maximum <code>IntegerNumber</code> that can be represented by this
    * class.
    */
   public final static LongIntegerNumber MAXIMUM_VALUE =
      new LongIntegerNumber(Long.MAX_VALUE);

   /**
    * Cached instance that represents the number minus two, -2.
    */
   public final static LongIntegerNumber MINUS_TWO =
      new LongIntegerNumber(-2);

   /**
    * Cached instance that represents the number minus one, -1.
    */
   public final static LongIntegerNumber MINUS_ONE =
      new LongIntegerNumber(-1);

   /**
    * Cached instance that represents the number zero, 0.
    */
   public final static LongIntegerNumber ZERO = new LongIntegerNumber(0);

   /**
    * Cached instance that represents the number one, 1.
    */
   public final static LongIntegerNumber ONE = new LongIntegerNumber(1);

   /**
    * Cached instance that represents the number two, 2.
    */
   public final static LongIntegerNumber TWO = new LongIntegerNumber(2);

   /**
    * Cached instance that represents the number three, 3.
    */
   public final static LongIntegerNumber THREE = new LongIntegerNumber(3);

   /**
    * Cached instance that represents the number four, 4.
    */
   public final static LongIntegerNumber FOUR = new LongIntegerNumber(4);

   /**
    * Cached instance that represents the number five, 5.
    */
   public final static LongIntegerNumber FIVE = new LongIntegerNumber(5);

   /**
    * Cached instance that represents the number six, 6.
    */
   public final static LongIntegerNumber SIX = new LongIntegerNumber(6);

   /**
    * Cached instance that represents the number seven, 7.
    */
   public final static LongIntegerNumber SEVEN = new LongIntegerNumber(7);

   /**
    * Cached instance that represents the number eight, 8.
    */
   public final static LongIntegerNumber EIGHT = new LongIntegerNumber(8);

   /**
    * Cached instance that represents the number nine, 9.
    */
   public final static LongIntegerNumber NINE = new LongIntegerNumber(9);

   /**
    * Cached instance that represents the number ten, 10.
    */
   public final static LongIntegerNumber TEN = new LongIntegerNumber(10);


   //=========================================================================
   // PROPERTY: value
   //=========================================================================

   /**
    * The <code>long</code> this <code>IntegerNumber</code> implementation
    * is based on.
    */
   public final long value;


   //=========================================================================
   // PROPERTY: sign
   //=========================================================================

   public int getSign()
   {
      if (value > 0)
      {
         return 1;
      }
      return (value == 0 ? 0 : -1);
   }


   //=========================================================================
   // PRIMARY UNARY OPERATIONS:
   //
   // - negate
   //
   //=========================================================================

   //------------------------------------------
   // METHOD OVERLOADED TO OPTIMIZE PERFORMANCE
   //------------------------------------------
   public RealNumber negate()
   {
      return createInstance(-value);
   }


   //=========================================================================
   // NARROWED PRIMARY BINARY OPERATIONS:
   //
   // - add
   // - multiply
   //
   //=========================================================================

   public IntegerNumber add(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      if (n.fitsLong())
      {
         long value2 = n.longValue();

         // both negative or both positive
         if ((value>=0 || value2>=0 || value>=(Long.MIN_VALUE - value2)) &&
             (value<=0 || value2<=0 || value<=(Long.MAX_VALUE - value2)))
         {
            return createInstance(value + value2);
         }
      }

      IntegerNumber big =
         BasicIntegerNumber.createInstance(BigInteger.valueOf(value));
      return big.add(n);
   }

   public IntegerNumber multiply(IntegerNumber n)
      throws IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      if (n.fitsLong())
      {
         long value2 = n.longValue();

         // zero result, value is zero
         if (value==0)
         {
            return this;
         }

         // zero result, value2 is zero
         if (value2==0)
         {
            return n;
         }

         // negative result
         if ((value>0 && value2<0 && value <= (Long.MIN_VALUE / value2)) ||
             (value<0 && value2>0 && value2 <= (Long.MIN_VALUE / value)))
         {
            return createInstance(value * value2);
         }

         // positive result
         if (value2 <= (Long.MAX_VALUE / value))
         {
            return createInstance(value * value2);
         }

         //-------------
         // FALL THROUGH
         //-------------
      }

      IntegerNumber big =
         BasicIntegerNumber.createInstance(BigInteger.valueOf(value));
      return big.multiply(n);
   }


   //=========================================================================
   // UNARY INTEGER OPERATIONS:
   //
   // - dec
   // - inc
   //
   //=========================================================================

   //------------------------------------------
   // METHOD OVERLOADED TO OPTIMIZE PERFORMANCE
   //------------------------------------------
   public IntegerNumber dec()
   {
      if (value > Long.MIN_VALUE)
      {
         return createInstance(value-1);
      }

      IntegerNumber big =
         BasicIntegerNumber.createInstance(BigInteger.valueOf(value));
      return big.dec();
   }

   //------------------------------------------
   // METHOD OVERLOADED TO OPTIMIZE PERFORMANCE
   //------------------------------------------
   public IntegerNumber inc()
   {
      if (value < Long.MAX_VALUE)
      {
         return createInstance(value+1);
      }

      IntegerNumber big =
         BasicIntegerNumber.createInstance(BigInteger.valueOf(value));
      return big.inc();
   }


   //=========================================================================
   // BINARY INTEGER OPERATIONS:
   //
   // - integerDivide
   // - remainder
   //
   //=========================================================================

   public IntegerNumber integerDivide(IntegerNumber n)
      throws ArithmeticException, IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      if (n.fitsLong())
      {
         long value2 = n.longValue();

         if (value2==0)
         {
            throw new ArithmeticException("divide by zero");
         }

         return createInstance(value / value2);
      }

      return ZERO;
   }

   public IntegerNumber remainder(IntegerNumber n)
      throws ArithmeticException, IllegalArgumentException
   {
      ExceptionSupport.checkNNotNull(n);

      if (n.fitsLong())
      {
         long value2 = n.longValue();

         if (value2==0)
         {
            throw new ArithmeticException("divide by zero");
         }

         return (IntegerNumber) createInstance(value % value2).abs();
      }

      return super.remainder(n);
   }

 
   //=========================================================================
   // NUMERIC CONVERSION:
   //
   // - longValue
   // - doubleValue
   // - toByteArray
   // - toBigInteger
   //
   //=========================================================================

   public long longValue()
   {
      return value;
   }

   public double doubleValue()
   {
      return (double) value;
   }

   public byte[] toByteArray()
   {
      //------------------
      // NOTE: inefficient
      //------------------

      return BigInteger.valueOf(value).toByteArray();
   }

   public BigInteger toBigInteger()
   {
      return BigInteger.valueOf(value);
   }


   //=========================================================================
   // TEXTUAL CONVERSION:
   //
   // - toString
   //
   //=========================================================================

   public String toString()
   {
      return String.valueOf(value);
   }

}
