/*
 * NumberCentral.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;

/**
 * A static factory that produces <code>RealNumber</code> objects.
 *
 * @version 1.0 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */
public final class NumberCentral
   extends Object
{
   //=========================================================================
   // CONSTRUCTOR
   //=========================================================================

   private NumberCentral()
   {
      // empty
   }


   //=========================================================================
   // STATIC PROPERTY: defaultFactory
   //
   // - defaultFactory
   // - getDefaultFactory
   // - setDefaultFactory
   //
   //=========================================================================

   /**
    * The default <code>NumberFactory</code>, which can be retrieved by
    * <code>getDefaultFactory()</code>.
    */
   static NumberFactory defaultFactory =
      DefaultNumberFactory.SHARED_INSTANCE;

   /**
    * Retrieves the default <code>NumberFactory</code>. This method will
    * never return <code>null</code>.
    *
    * @return
    *    the <code>NumberFactory</code> that is set as the default.
    */
   public final static NumberFactory getDefaultFactory()
   {
      return defaultFactory;
   }

   /**
    * Sets the default <code>NumberFactory</code>.
    *
    * <p>The default <code>NumberFactory</code> will be set to the
    * fallback default if the argument is <code>null</code>.
    *
    * @param factory
    *    the new default factory, or <code>null</code> if the
    *    default factory should be set to the fallback default.
    */
   public final static void setDefaultFactory(NumberFactory newDefault)
   {
      defaultFactory =
         newDefault != null ? newDefault
                            : DefaultNumberFactory.SHARED_INSTANCE;
   }


   //=========================================================================
   // FACTORY METHODS FOR COMPOSITES:
   //
   // - add
   // - multiply
   // - pow
   //
   //=========================================================================

   /**
    * Computes the sum of the arguments.
    *
    * @param a the first operand
    * @param b the second operand
    * @return the sum of <em>a</em> and <em>b</em>.
    * @exception IllegalArgumentException if either argument is
    *    <code>null</code>.
    */
   public static RealNumber add(RealNumber a,
                                RealNumber b)
      throws IllegalArgumentException
   {
      return defaultFactory.add(a, b);
   }

   /**
    * Computes the product of the arguments.
    *
    * @param a the first operand
    * @param b the second operand
    * @return the product of <em>a</em> and <em>b</em>.
    * @exception IllegalArgumentException if either argument is
    *    <code>null</code>.
    */
   public static RealNumber multiply(RealNumber a,
                                     RealNumber b)
      throws IllegalArgumentException
   {
      return defaultFactory.multiply(a, b);
   }

   /**
    * Computes the a power for the given base and exponent.
    *
    * @param base the base for the power.
    * @param exponent the exponent for the power.
    * @return <em>base</em><sup><em>exponent</em></sup>
    * @exception IllegalArgumentException if either argument is
    *    <code>null</code>.
    * @exception ArithmeticException if the power cannot be computed.
    */
   public static RealNumber pow(RealNumber base,
                                RealNumber exponent)

      throws IllegalArgumentException,
             ArithmeticException
   {
      return defaultFactory.pow(base, exponent);
   }


   //=========================================================================
   // FACTORY METHODS:
   //
   // - createFraction (2x)
   // - createInteger
   //
   //=========================================================================

   /**
    * Returns a fraction with the given numerator and denominator, both
    * being a <code>long</code>.
    *
    * @param numerator the numerator for the fraction.
    * @param denominator the denominator for the fraction.
    * @return the (possibly newly constructed) fraction.
    * @exception ArithmeticException if the value of <em>denominator</em> is
    *    zero.
    */
   public static RationalNumber createFraction(long numerator,
                                               long denominator)

      throws ArithmeticException
   {
      return defaultFactory.createFraction(numerator,
                                                         denominator);
   }

   /**
    * Returns a fraction with the given numerator and denominator.
    *
    * @param numerator the numerator for the fraction.
    * @param denominator the denominator for the fraction.
    * @return the (possibly newly constructed) fraction.
    * @exception ArithmeticException if the value of <em>denominator</em> is
    *    zero.
    * @exception IllegalArgumentException if either one or both of the
    *    arguments are <code>null</code>.
    */
   public static RationalNumber createFraction(IntegerNumber numerator,
                                               IntegerNumber denominator)

      throws ArithmeticException,
             IllegalArgumentException
   {
      return defaultFactory.createFraction(numerator,
                                                         denominator);
   }


   /**
    * Returns an integer number from a byte array in two's complement
    * notation.
    *
    * @param bytes the two's complement byte array.
    * @return the (possibly newly constructed) integer number.
    * @exception IllegalArgumentException if the argument is
    *    <code>null</code> or the array has length 0.
    */
   public static IntegerNumber createInteger(byte[] bytes)
      throws IllegalArgumentException
   {
      return defaultFactory.createInteger(bytes);
   }

   //---------------------------------------------
   // PENDING: array length==0 --> return value 0?
   //---------------------------------------------


   //=========================================================================
   // valueOf() methods
   //
   // - double
   // - long
   // - BigInteger
   // - BigDecimal
   //
   //=========================================================================

   /**
    * Returns a rational number from the given <code>double</code>.
    *
    * @param n the original value.
    * @return the converted <code>RealNumber</code>.
    * @exception IllegalArgumentException if the object represents Infinity,
    *    -Infinity or NaN.
    */
   public static RationalNumber valueOf(double n)
      throws IllegalArgumentException
   {
      return defaultFactory.valueOf(n);
   }

   /**
    * Returns a rational number from the given <code>long</code>.
    *
    * @param n the original value.
    * @return the converted <code>RealNumber</code>.
    */
   public static IntegerNumber valueOf(long n)
   {
      return defaultFactory.valueOf(n);
   }

   /**
    * Converts a <code>java.math.BigInteger</code> to an
    * <code>IntegerNumber</code>.
    *
    * @param n the <code>BigInteger</code> to convert.
    * @return the (possibly newly constructed) integer number with the
    *    value of <em>n</em>.
    * @exception IllegalArgumentException if the argument is
    *    <code>null</code>.
    */
   public static IntegerNumber valueOf(BigInteger n)
      throws IllegalArgumentException
   {
      return defaultFactory.valueOf(n);
   }

   /**
    * Converts a <code>java.math.BigDecimal</code> to an
    * <code>IntegerNumber</code>.
    *
    * @param n the <code>BigDecimal</code> to convert.
    * @return the (possibly newly constructed) rational number with the
    *    value of <em>n</em>.
    * @exception IllegalArgumentException if the argument is
    *    <code>null</code>.
    */
   public static RationalNumber valueOf(BigDecimal n)
      throws IllegalArgumentException
   {
      return defaultFactory.valueOf(n);
   }

   /**
    * Returns an <code>IntegerNumber</code> with a random value.
    *
    * @param the maximum number of bytes for the constructed number.
    * @return a random integer.
    */
   public static IntegerNumber createRandomInteger(int numBytes)
   {
      return defaultFactory.createRandomInteger(numBytes);
   }

}
