/*
 * NumberFactory.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;
import java.util.*;

/**
 * Interface for a factory that produces <code>RealNumber</code> objects.
 *
 * @version 1.2 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */
public interface NumberFactory
{
   //=========================================================================
   // FACTORY METHODS FOR COMPOSITES:
   //
   // - add
   // - multiply
   // - pow
   //
   //=========================================================================

   /**
    * Computes the sum of the arguments.
    *
    * @param a the first operand.
    * @param b the second operand.
    * @return  the sum of <em>a</em> and <em>b</em>.
    *
    * @exception IllegalArgumentException
    *    if one of the following applies:
    *    <ol>
    *       <li><em>a</em> == <code>null</code></li>
    *       <li><em>b</em> == <code>null</code></li>
    *    </ol>
    */
   public RealNumber add(RealNumber a,
                         RealNumber b)
      throws IllegalArgumentException;

   /**
    * Computes the product of the arguments.
    *
    * @param a the first operand.
    * @param b the second operand.
    * @return  the product of <em>a</em> and <em>b</em>.
    *
    * @exception IllegalArgumentException
    *    if one of the following applies:
    *    <ol>
    *       <li><em>a</em> == <code>null</code></li>
    *       <li><em>b</em> == <code>null</code></li>
    *    </ol>
    */
   public RealNumber multiply(RealNumber a,
                              RealNumber b)
      throws IllegalArgumentException;

   /**
    * Computes the power for the given base and exponent.
    *
    * @param base     the base for the power.
    * @param exponent the exponent for the power.
    * @return         <em>base</em><sup><em>exponent</em></sup>
    *
    * @exception IllegalArgumentException
    *    if one of the following applies:
    *    <ol>
    *       <li><em>base</em>     == <code>null</code></li>
    *       <li><em>exponent</em> == <code>null</code></li>
    *    </ol>
    *
    * @exception ArithmeticException
    *    if the power cannot be computed.
    */
   public RealNumber pow(RealNumber base,
                         RealNumber exponent)
      throws IllegalArgumentException,
             ArithmeticException;


   //=========================================================================
   // FACTORY METHODS:
   //
   // - createFraction (2x)
   // - createInteger
   // - valueOf (4x)
   //
   //=========================================================================

   /**
    * Returns a fraction with the given numerator and denominator, both
    * being a <code>long</code>.
    *
    * @param numerator   the numerator for the fraction.
    * @param denominator the denominator for the fraction.
    * @return            the (possibly newly constructed) fraction.
    *
    * @exception ArithmeticException
    *    if the value of <em>denominator</em> is zero.
    */
   public RationalNumber createFraction(long numerator,
                                        long denominator)
      throws ArithmeticException;

   /**
    * Returns a fraction with the given numerator and denominator.
    *
    * @param numerator   the numerator for the fraction.
    * @param denominator the denominator for the fraction.
    * @return            the (possibly newly constructed) fraction.
    *
    * @exception IllegalArgumentException
    *    if one of the following applies:
    *    <ol>
    *       <li><em>numerator</em>   == <code>null</code></li>
    *       <li><em>denominator</em> == <code>null</code></li>
    *    </ol>
    *
    * @exception ArithmeticException
    *    if the value of <em>denominator</em> is zero.
    */
   public RationalNumber createFraction(IntegerNumber numerator,
                                        IntegerNumber denominator)
      throws IllegalArgumentException,
             ArithmeticException;

   /**
    * Returns an integer number from a byte array in two's complement
    * notation.
    *
    * @param bytes the two's complement byte array.
    * @return      the (possibly newly constructed) integer number.
    *
    * @exception IllegalArgumentException
    *    if one of the following applies:
    *    <ol>
    *       <li><em>bytes</em> is <code>null</code></li>
    *       <li>the size of <em>bytes</em> is 0.
    *    </ol>
    */
   public IntegerNumber createInteger(byte[] bytes)
      throws IllegalArgumentException;

   /**
    * Returns a rational number from the given <code>double</code>.
    *
    * @param n the original value.
    * @return  the converted <code>RealNumber</code>.
    *
    * @exception IllegalArgumentException
    *    if any of the following applies:
    *    <ol>
    *       <li><em>n</em> represents Infinity</li>
    *       <li><em>n</em> represents -Infinity</li>
    *       <li><em>n</em> represents NaN</li>
    *    </ol>
    */
   public RationalNumber valueOf(double n)
      throws IllegalArgumentException;

   /**
    * Returns a rational number from the given <code>long</code>.
    *
    * @param n the original value.
    * @return the converted <code>RealNumber</code>.
    */
   public IntegerNumber valueOf(long n);

   /**
    * Converts a <code>java.math.BigInteger</code> to an
    * <code>IntegerNumber</code>.
    *
    * @param n the <code>BigInteger</code> to convert.
    * @return  the (possibly newly constructed) integer number with the
    *          value of <em>n</em>.
    *
    * @exception IllegalArgumentException
    *    if <em>n</em> == <code>null</code>.
    */
   public IntegerNumber valueOf(BigInteger n)
      throws IllegalArgumentException;

   /**
    * Converts a <code>java.math.BigDecimal</code> to a
    * <code>RationalNumber</code>.
    *
    * @param n the <code>BigDecimal</code> to convert.
    * @return  the (possibly newly constructed) rational number with the
    *          value of <em>n</em>.
    *
    * @exception IllegalArgumentException
    *    if the argument is <code>null</code>.
    */
   public RationalNumber valueOf(BigDecimal n)
      throws IllegalArgumentException;

   /**
    * Converts a <code>String</code> to a <code>RealNumber</code>.
    *
    * @param s the string to parse.
    * @return  the constructed <code>RealNumber</code>, on success.
    *
    * @exception NumberFormatException
    *    if <em>s</em> cannot be parsed.
    */
   public RealNumber valueOf(String s)
      throws IllegalArgumentException,
             NumberFormatException;


   //=========================================================================
   // RANDOM NUMBERS
   //=========================================================================

   /**
    * Returns a pseudo-random <code>IntegerNumber</code> of at maximum
    * <code>maxBytes</code> bytes. The returned integer number will have a
    * value between -2<sup><em>maxBytes</em>*8</sup> and
    * (2<sup><em>maxBytes</em></sup>-1).
    *
    * @return the pseudo-random integer.
    *
    * @exception IllegalArgumentException
    *    if <em>maxBytes</em> &lt; 1.
    */
   public IntegerNumber createRandomInteger(int maxBytes)
      throws IllegalArgumentException;

}
