/*
 * Product.java
 *
 * Copyright (C) 1998  Utrecht University, Netherlands
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package nl.coffee.math;

import java.math.*;

/**
 * A product of two real numbers.
 *
 * @version 1.1 199809.24
 * @author Ernst de Haan
 *
 * @since JUMP_1.0
 */
public class Product
   extends AbstractCompositeNumber
{
   //=========================================================================
   // CONSTRUCTOR
   //=========================================================================

   /**
    * Constructs a <code>Product</code> based on the 2 specified operands.
    *
    * @param operand1 the first operand for the product.
    * @param operand2 the second operand for the product.
    * @exception IllegalArgumentException if either argument
    *    is <code>null</code>.
    */
   protected Product(RealNumber a, RealNumber b)
      throws IllegalArgumentException
   {
      this.elements = new RealNumber[2];
      elements[0] = a;
      elements[1] = b;

      sign = determineSign(a, b);
   }


   //=========================================================================
   // FACTORY METHOD
   //=========================================================================

   /**
    * Returns a <code>Product</code> with the specified operands.
    *
    * @param a the first operand.
    * @param b the second operand.
    * @return the <code>Product</code> instance, possibly newly constructed.
    * @exception IllegalArgumentException if either argument is
    *    <code>null</code>.
    */
   public static Product createInstance(RealNumber a, RealNumber b)
      throws IllegalArgumentException
   {
      return new Product(a, b);
   }


   //=========================================================================
   // DETERMINE SIGN OF SUM:
   //
   // - determineSign
   //
   //=========================================================================

   /**
    * Computes the sign of a product with the specified operands.
    *
    * @param operand1 the first operand.
    * @param operand2 the second operand.
    */
   protected static int determineSign(RealNumber a, RealNumber b)
   {
      return (a.getSign() * b.getSign());
   }


   //=========================================================================
   // PROPERTY: sign
   //
   // - sign
   // - getSign
   //
   //=========================================================================

   /**
    * The sign of this product. This field is set only once upon construction.
    */
   protected int sign;

   /**
    * Determines the sign of this number. This method returns -1, 0, or
    * 1 as this number is negative, zero, or positive.
    *
    * @return the sign of this number.
    */
   public int getSign()
   {
      return sign;
   }


   //=========================================================================
   // COMPARISON:
   //
   // - retryCompare
   //
   //=========================================================================

   /**
    * Returns -1, 0 or 1 as this number is less than, equal to, or greater
    * than <em>n</em>, where <em>n</em> is a real number.
    *
    * @param n the number to compare to.
    * @return -1 if this&lt;<em>n</em>, 0 if this==<em>n</em>, 1 if
    *    this&gt;<em>n</em>.
    * @exception IllegalArgumentException if the argument is
    *    <code>null</code>.
    */
   public int retryCompare(RealNumber n)
      throws IllegalArgumentException,
             CanNotCompareException
   {
      //------------------------------
      // PENDING: How do we do this ?!
      //------------------------------

      throw new CanNotCompareException(n, this);
   }


   //=========================================================================
   // NUMERIC CONVERSION
   //
   // - toBigDecimal
   //
   //=========================================================================

   /**
    * Converts the value of this number to a <code>BigDecimal</code> with the
    * specified precision and rounding mode.
    *
    * @param precision the number of digits behind the decimal point.
    * @param roundingMode the rounding mode to use, one of the modes defined
    *    in class <code>BigDecimal</code>.
    * @return a <code>BigDecimal</code> with the rounded value of this.
    * @exception IllegalArgumentException if <em>precision</em>&lt;0 or
    *    the rounding mode is not one of the valid rounding modes defined in
    *    class <code>BigDecimal</code>.
    */
   public BigDecimal toBigDecimal(int precision, int roundingMode)
      throws IllegalArgumentException
   {
      BigDecimal a = elements[0].toBigDecimal(precision+1, roundingMode);
      BigDecimal b = elements[1].toBigDecimal(precision+1, roundingMode);

      // compute BigDecimal product
      BigDecimal result = a.multiply(b);

      // return correct precision
      return result.setScale(precision, roundingMode);
   }


   //=========================================================================
   // ROUNDING:
   //
   // - trunc
   //
   //=========================================================================

   /**
    * Rounds to an integer number towards 0.
    *
    * @return this number truncated to an integer.
    */
   public IntegerNumber trunc()
   {
      //-------------------------------------
      // PENDING: Reconsider this
      //          This is pretty safe for now
      //-------------------------------------

      return NumberCentral.valueOf(toBigDecimal(0)).trunc();
   }


   //=========================================================================
   // TEXTUAL CONVERSION:
   //
   // - toString
   //
   //=========================================================================

   /**
    * Returns a textual representation of this object.
    *
    * @return a <code>String</code> representation of this object.
    */
   public String toString()
   {
      StringBuffer buffer = new StringBuffer();

      if (elements[0] instanceof CompositeNumber)
      {
         buffer.append('(');
         buffer.append(elements[0]);
         buffer.append(")*");
      }
      else
      {
         buffer.append(elements[0]);
         buffer.append('*');
      }

      if (elements[1] instanceof CompositeNumber)
      {
         buffer.append('(');
         buffer.append(elements[1]);
         buffer.append(')');
      }
      else
      {
         buffer.append(elements[1]);
      }

      return buffer.toString();
   }

}
